from clubcpg.ParseBam import BamFileReadParser
import os
import logging
from multiprocessing import Pool
import numpy as np
from collections import defaultdict
import time
from pandas.core.indexes.base import InvalidIndexError


class CalculateCompleteBins:
    """
    Class to calculate the number of reads covering all CpGs
    """
    def __init__(self,
                 bam_file, bin_size, output_directory, number_of_processors=1, mbias_read1_5=None, mbias_read1_3=None, mbias_read2_5= None, mbias_read2_3=None, no_overlap=True):
        """
        This class is initialized with a path to a bam file and a bin size
    
        :param bam_file: One of the BAM files for analysis to be performed
        :param bin_size: Size of the bins for the analysis, integer
        :number_of_processors: How many CPUs to use for parallel computation, default=1
        """
        self.input_bam_file = bam_file
        self.bin_size = int(bin_size)
        self.number_of_processors = int(number_of_processors)
        self.output_directory = output_directory
        self.bins_no_reads = 0
        self.bins_yes_reads = 0
    
        self.mbias_read1_5 = mbias_read1_5
        self.mbias_read1_3 = mbias_read1_3
        self.mbias_read2_5 = mbias_read2_5
        self.mbias_read2_3 = mbias_read2_3
        self.no_overlap = no_overlap

    def calculate_bin_coverage(self, bin):
        """
        Take a single bin, return a matrix. This is passed to a multiprocessing Pool.

        :param bin: Bin should be passed as "Chr19_4343343"
        :return: pd.DataFrame with rows containing NaNs dropped
        """
        # Get reads from bam file
        parser = BamFileReadParser(self.input_bam_file, 20, self.mbias_read1_5, self.mbias_read1_3,
                                   self.mbias_read2_5, self.mbias_read2_3, self.no_overlap)
        # Split bin into parts
        chromosome, bin_location = bin.split("_")
        bin_location = int(bin_location)
        try:
            reads = parser.parse_reads(chromosome, bin_location-self.bin_size, bin_location)
            matrix = parser.create_matrix(reads)
        except BaseException as e:
            # No reads are within this window, do nothing
            self.bins_no_reads += 1
            return None
        except:
            logging.error("Unknown error: {}".format(bin))
            return None

        # drop rows of ALL NaN
        matrix = matrix.dropna(how="all")
        # convert to data_frame of 1s and 0s, drop rows with NaN
        matrix = matrix.dropna()
        # if matrix is empty, attempt to create it with correction before giving up
        if len(matrix) == 0:
            original_matrix = matrix.copy()
            reads = parser.correct_cpg_positions(reads)
            try:
                matrix = parser.create_matrix(reads)
            except InvalidIndexError as e:
                logging.error("Invalid Index error when creating matrices at bin {}".format(bin))
                logging.debug(str(e))
                return bin, original_matrix
            except ValueError as e:
                logging.error("Matrix concat error ar bin {}".format(bin))
                logging.debug(str(e))

            matrix = matrix.dropna()
            if len(matrix) > 0:
                logging.info("Correction attempt at bin {}: SUCCESS".format(bin))
            else:
                logging.info("Correction attempt at bin {}: FAILED".format(bin))

        return bin, matrix

    def get_chromosome_lengths(self):
        """
        Get dictionary containing lengths of the chromosomes. Uses bam file for reference

        :return: Dictionary of chromosome lengths, ex: {"chrX": 222222}
        """
        parser = BamFileReadParser(self.input_bam_file, 20)
        return dict(zip(parser.OpenBamFile.references, parser.OpenBamFile.lengths))

    @staticmethod
    def remove_scaffolds(chromosome_len_dict):
        """
        Return a dict containing only the standard chromosomes starting with "chr"

        :param chromosome_len_dict: A dict generated by get_chromosome_lenghts()
        :return: a dict containing only chromosomes starting with "chr"
        """
        new_dict = dict(chromosome_len_dict)
        for key in chromosome_len_dict.keys():
            if not key.startswith("chr"):
                new_dict.pop(key)

        return new_dict

    def generate_bins_list(self, chromosome_len_dict):
        """
        Get a dict of lists of all bins according to desired bin size for all chromosomes in the passed dict

        :param chromosome_len_dict: A dict of chromosome length sizes from get_chromosome_lenghts, cleaned up by remove_scaffolds() if desired
        :return: dict with each key being a chromosome. ex: chr1
        """
        all_bins = defaultdict(list)
        for key, value in chromosome_len_dict.items():
            bins = list(np.arange(self.bin_size, value + self.bin_size, self.bin_size))
            bins = ["_".join([key, str(x)]) for x in bins]
            all_bins[key].extend(bins)

        return all_bins

    def analyze_bins(self, individual_chrom=None):
        """
        Main function in class. Run the Complete analysis on the data

        :param individual_chrom: Chromosome to analyze: ie "chr7"
        :return: filename of the generated report
        """

        # Track the progress of the multiprocessing and output
        def track_progress(job, update_interval=60):
            while job._number_left > 0:
                logging.info("Tasks remaining = {0}".format(
                    job._number_left * job._chunksize))
                time.sleep(update_interval)

        # Get and clean dict of chromosome lenghts, convert to list of bins
        chromosome_lengths = self.get_chromosome_lengths()
        chromosome_lengths = self.remove_scaffolds(chromosome_lengths)

        # If one chromosome was specified use only that chromosome
        if individual_chrom:
            new = dict()
            new[individual_chrom] = chromosome_lengths[individual_chrom]
            chromosome_lengths = new

        bins_to_analyze = self.generate_bins_list(chromosome_lengths)

        # Set up for multiprocessing
        # Loop over bin dict and pool.map them individually
        final_results = []
        for key in bins_to_analyze.keys():
            pool = Pool(processes=self.number_of_processors)
            results = pool.map_async(self.calculate_bin_coverage, bins_to_analyze[key])

            track_progress(results)

            # once done, get results
            results = results.get()

            final_results.extend(results)

        logging.info("Analysis complete")

        # Write to output file
        output_file = os.path.join(self.output_directory, "CompleteBins.{}.{}.csv".format(os.path.basename(self.input_bam_file), individual_chrom))

        with open(output_file, "w") as out:
            for result in final_results:
                if result:
                    # bin
                    out.write(result[0] + ",")
                    # num of reads
                    out.write(str(result[1].shape[0]) + ",")
                    # num of CpGs
                    out.write(str(result[1].shape[1]) + "\n")

        logging.info("Full read coverage analysis complete!")
        return output_file



from __future__ import print_function

"""
Author: 
Jack Duryea
Waterland Lab
Computational Epigenetics Section
Baylor College of Medicine

Created April 2018

Updated April 11 2019: use random forests as model


PReLIM: Preceise Read Level Imputation of Methylation

PReLIM imputes missing CpG methylation
states in CpG matrices.

"""

# standard imports
import numpy as np
from tqdm import tqdm
import copy
import time

from collections import defaultdict
import random

# sklearn imports
from sklearn.preprocessing import normalize
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import GridSearchCV

# Pickle
try:
	import cPickle as p
except ImportError:
	import pickle as p


# TODO: most of these fields are redundant in our application
class CpGBin():
	""" 
	Constructor for a bin

	"""
	def __init__(self, 
			matrix, 
			#relative_positions
			binStartInc=None, 
			binEndInc=None, 
			cpgPositions=None, 
			sequence="",
			encoding=None, 
			missingToken= -1, 
			chromosome=None, 
			binSize=100, 
			species="MM10", 
			verbose=True, 
			tag1=None, 
			tag2=None):
		"""

		:param matrix: numpy array, the bin's CpG matrix.
		:param binStartInc: integer, the starting, inclusive, chromosomal index of the bin.
		:param binEndInc: integer, the ending, inclusive, chromosomal index of the bin.
		:param cpgPositions: array of integers, the chromosomal positions of the CpGs in the bin.
		:param sequence: string, nucleotide sequence (A,C,G,T)
		:param encoding: array, a reduced representation of the bin's CpG matrix
		:param missingToken: integer, the token that represents missing data in the matrix.
		:param chromosome: string, the chromosome this bin resides in.
		:param binSize: integer, the number of base pairs this bin covers
		:param species: string, the speices this bin belongs too.
		:param verbose: boolean, print warnings, set to "false" for no error checking and faster speed
		:param tag1: anything, for custom use.
		:param tag2: anything, for custom use.

		"""


		self.cpgDensity = matrix.shape[1]
		self.readDepth = matrix.shape[0]
		
		

		self.matrix = np.array(matrix, dtype=float)
		self.binStartInc = binStartInc
		self.binEndInc = binEndInc
		self.cpgPositions = cpgPositions
		self.sequence = sequence
		self.missingToken = missingToken
		self.chromosome = chromosome
		self.binSize = binSize
		self.species = species
		self.tag1 = tag1
		self.tag2 = tag2


class PReLIM():
	"""
	PReLIM imputation class to handle training and predicting from models.

	"""
	def __init__(self, cpgDensity=2):
		self.model = None
		self.cpgDensity = cpgDensity
		self.METHYLATED = 1
		self.UNMETHYLATED = 0
		self.MISSING = -1
		self.methylated = 1
		self.unmethylated = 0
		self.unknown = -1

	# Train a model
	def train(self, bin_matrices, model_file="no", verbose=False):
		"""
		bin_matrices: list of cpg matrices

		model_file, string,      The name of the file to save the model to. 
			If None, then create a file name that includes a timestamp.
			If you don't want to save a file, set this to "no"
		"""
		# bin_matrices: a list of cpg matrices 
		X,y = self.get_X_y(bin_matrices, model_file=model_file, verbose=False)
		
		# Train the neural network model
		self.fit(X,y, model_file=model_file, verbose=verbose)

	def fit(self,
			X_train,
			y_train,
			n_estimators = [10, 50, 100, 500, 1000],
			cores = -1,
			max_depths = [1, 5, 10, 20, 30],
			model_file=None,
			verbose=False
			):
		"""
		Inputs: 
		1. X_train,     numpy array, Contains feature vectors.
		2. y_train,     numpy array, Contains labels for training data.
		3. n_estimators, list, the number of estimators to try during a grid search.
		4. max_depths, list, the maximum depths of trees to try during a grid search.
		5. cores, the number of cores to use during training, helpful for grid search.
		6. model_file, string,      The name of the file to save the model to. 
			If None, then create a file name that includes a timestamp.
			If you don't want to save a file, set this to "no"
	
		5-fold validation is built into the grid search

		Outputs: 
		The trained model

		Usage: 
		model.fit(X_train, y_train)	
		"""


		grid_param = {  
		 "n_estimators": n_estimators,
		  "max_depth": max_depths,
		}

		# Note: let the grid search use a lot of cores, but only use 1 for each forest
		# since dispatching can take a lot of time
		rf = RandomForestClassifier(n_jobs=1)
		self.model = GridSearchCV(rf, grid_param, n_jobs=cores, cv=5, verbose=verbose)
		self.model.fit(X_train, y_train)


		# save the model

		if model_file == "no":
			return self.model

		if not model_file:
			model_file = "PReLIM_model" + str(time.time())

		p.dump(self.model, open(model_file,"wb"))

		return self.model




	# Feature collection directly from bins
	def get_X_y(self, bin_matrices, model_file=None, verbose=False):
		bins = []

		# convert to bin objects for ease of use
		for matrix in bin_matrices:
			mybin = CpGBin( matrix=matrix )
			bins.append( mybin )
		
		# find bins with no missing data
		complete_bins = _filter_missing_data( bins )
		random.shuffle( complete_bins )
		
		# apply masks
		masked_bins = _apply_masks( complete_bins, bins )

		# extract features
		X, y = self._collectFeatures( masked_bins ) 
		return X, y 


	# Return a vector of predicted classes 
	def predict_classes(self, X):
		"""
		Inputs: 
		1. X, numpy array, contains feature vectors
		
		Outputs: 
		1. 1-d numpy array of prediction values

		Usage: 
		y_pred = CpGNet.predict_classes(X)  

		"""
		return self.model.predict(X)
	
	# Return a vector of probabilities for methylation
	def predict(self, X):
		"""
		Inputs: 
		1. X, numpy array, contains feature vectors
		
		Outputs: 
		1. 1-d numpy array of predicted class labels

		Usage: 
		y_pred = CpGNet.predict(X)  

		"""
		return self.model.predict_proba(X)[:,1]


	def predict_proba(self, X):
		"""
		Inputs: 
		1. X, numpy array, contains feature vectors
		
		Outputs: 
		1. 1-d numpy array of class predictions 

		Usage: 
		y_pred = CpGNet.predict(X)  

		"""
		return self.model.predict_proba(X)[:1]


	# Load a saved model
	def loadWeights(self, model_file):
		"""
		Inputs:
		1. model_file, string, name of file with a saved model

		Outputs:
		None

		Effects:
		self.model is loaded with the provided weights
		"""
		self.model = p.load(open(model_file,"rb"))

	




	def _get_imputation_features(self,matrix):
		'''
		Returns a vector of features needed for the imputation of this matrix

		Inputs: 
		1. matrix, a 2d np array, dtype=float, representing a CpG matrix, 1=methylated, 0=unmethylated, -1=unknown

		Outputs:
		1. A feature vector for the matrix
		'''

		X = []

		numReads = matrix.shape[0]
		density = matrix.shape[1]

		nan_copy = np.copy(matrix)
		nan_copy[nan_copy == -1] = np.nan
		column_means = np.nanmean(nan_copy, axis=0)
		row_means = np.nanmean(nan_copy, axis=1)
		
		encoding = self._encode_input_matrix(matrix)[0]

		for i in range(numReads):
			for j in range(density):
				observed_state = matrix[i, j]
				
				if observed_state != -1:
					continue

				row_mean = row_means[i]
				col_mean = column_means[j]
				
				row = np.copy(matrix[i])
				row[j] = -1

				data = [row_mean] + [col_mean] +  [i, j] + list(row) +  list(encoding)
				X.append(data)

		X = np.array(X)

		return X

	# Imputes missing values in Bins
	def impute(self, matrix):
		"""
		Inputs: 
		1. matrix, a 2d np array, dtype=float, representing a CpG matrix, 1=methylated, 0=unmethylated, -1=unknown
		
		Outputs: 
		1. A 2d numpy array with predicted probabilities of methylation

		"""

		X = self._get_imputation_features(matrix)
		
		if len(X) == 0: # nothing to impute
			return matrix

		predictions = self.predict(X)

		k = 0 # keep track of prediction index for missing states
		predicted_matrix = np.copy(matrix)
		for i in range(predicted_matrix.shape[0]):
			for j in range(predicted_matrix.shape[1]):
				if predicted_matrix[i, j] == -1:
					predicted_matrix[i, j] = predictions[k]
					k += 1

		return predicted_matrix







	def impute_many(self, matrices):
		'''
		Imputes a bunch of matrices at the same time to help speed up imputation time.

		Inputs:

		1. matrices: array-like (i.e. list), where each element is
		a 2d np array, dtype=float, representing a CpG matrix, 1=methylated, 0=unmethylated, -1=unknown

		Outputs:

		1. A List of 2d numpy arrays with predicted probabilities of methylation for unknown values.
		'''

		# Extract all features for all matrices so we can predict in bulk, this is where the speedup comes from
		
		X = np.array([features for matrix_features in [self._get_imputation_features(matrix) for matrix in matrices] for features in matrix_features])
		
		if len(X) == 0:
			return matrices
		
		predictions = self.predict(X)


		predicted_matrices = []

		# TODO: lots of for-loops here, could be sped up?

		k = 0 # keep track of prediction index for missing states, order is crucial!
		for matrix in matrices:
			predicted_matrix = np.copy(matrix)
			for i in range(predicted_matrix.shape[0]):
				for j in range(predicted_matrix.shape[1]):
					if predicted_matrix[i, j] == -1:
						predicted_matrix[i, j] = predictions[k]
						k += 1
			predicted_matrices.append(predicted_matrix)

		return predicted_matrices







	### Helper functions, for private use only ###

	# Returns a matrix encoding of a CpG matrix
	def _encode_input_matrix(self, m):
		matrix = np.copy(m)
		n_cpgs = matrix.shape[1]
		matrix += 1  # deal with -1s
		base_3_vec = np.power(3, np.arange(n_cpgs - 1, -1, -1))
		#
		encodings = np.dot(base_3_vec, matrix.T)
		#
		encoded_vector_dim = np.power(3, n_cpgs)
		encoded_vector = np.zeros(encoded_vector_dim)
		#
		for x in encodings:
			encoded_vector[int(x)] += 1
		#
		num_reads = encodings.shape[0]
		#
		# Now we normalize
		encoded_vector_norm = normalize([encoded_vector], norm="l1")
		return encoded_vector_norm[0], num_reads

	# finds the majority class of the given column, discounting the current cpg
	
	def _get_column_mean(self, matrix, col_i, current_cpg_state):
		sub = matrix[:, col_i]
		return self._get_mean(sub, current_cpg_state)

	# finds the majority class of the given read, discounting the current cpg
	def _get_read_mean(self, matrix, read_i, current_cpg_state):
		sub = matrix[read_i, :]
		return self._get_mean(sub, current_cpg_state)

	def _get_mean(self, sub_matrix, current_cpg_state):
		num_methy = np.count_nonzero(sub_matrix == self.METHYLATED)
		num_unmethy = np.count_nonzero(sub_matrix == self.UNMETHYLATED)

		if current_cpg_state == self.METHYLATED:
			num_methy -= 1
		num_methy = max(0, num_methy)
		if current_cpg_state == self.UNMETHYLATED:
			num_unmethy -= 1
		num_unmethy = max(0, num_unmethy)
		if float(num_methy + num_unmethy) == 0:
			return -2

		return float(num_methy) / float(num_methy + num_unmethy)

	# Returns X, y
	# note: y can contain the labels 1,0, -1
	def _collectFeatures(self, bins):
		X = []
		Y = []
		for Bin in tqdm(bins):
			observed_matrix = Bin.tag2["observed"]
			truth_matrix = Bin.tag2["truth"]
			encoding = self._encode_input_matrix(observed_matrix)[0]

			numReads = observed_matrix.shape[0]
			density = observed_matrix.shape[1]
			#positions = Bin.cpgPositions
			nan_copy = np.copy(observed_matrix)
			

			nan_copy[nan_copy == -1] = np.nan 
			column_means = np.nanmean(nan_copy,axis=0)
			row_means = np.nanmean(nan_copy,axis=1)

			for i in range(numReads):
				for j in range(density):
					observed_state = observed_matrix[i,j]
					if observed_state != -1:
						continue

					state = truth_matrix[i,j]
					Y.append(state)
					
					
					
					row_mean = row_means[i]
					col_mean = column_means[j]
					# j is the current index in the row
					
					# encoding is the matrix encoding vector
					# differences is the difference in positions of the cpgs
					row = np.copy(observed_matrix[i])
					row[j] = -1

					data = [row_mean] + [col_mean] +  [i, j] + list(row) +  list(encoding)
					X.append(data)


		X = np.array(X)
		Y = np.array(Y)
		Y.astype(int)
		return X, Y



# Helper functions

# returns a list of bins similar to the input
# but matrix rows with missing values are removed
def _filter_bad_reads(bins):
	filtered_bins = []
	for Bin in bins:
		newBin = copy.deepcopy(Bin)
		matrix = newBin.matrix

		# find rows with missing values
		counts = np.count_nonzero(matrix == -1, axis=1)
		idx = counts == 0
		matrix_filtered = matrix[idx]
		newBin.matrix = matrix_filtered
		filtered_bins.append(newBin)

	return filtered_bins

# returns a mapping of dimensions to list of masks that can be used on data
# of that size.
# the missing pattern is in matrix form.
# -1 is missing, 2 is known
def _extract_masks( bins):
	masks = defaultdict(lambda: [])
	for Bin in tqdm(bins):
		matrix = np.copy(Bin.matrix)
		matrix[matrix >= 0] = 2

		#min_missing = 10
		min_missing = 1 # must have at least 1 missing value
		if np.count_nonzero(matrix == -1) >= min_missing:
			masks[matrix.shape].append(matrix)

	return masks

  
def _apply_masks( filtered_bins, all_bins ):
	
	masks = _extract_masks( all_bins )
	ready_bins = []

	for Bin in filtered_bins:
		truth_matrix = Bin.matrix
		m_shape = truth_matrix.shape
		if m_shape in masks:
			if len( masks [ m_shape ] ) > 0:
				mask = random.choice(masks[m_shape])
				observed = np.minimum(truth_matrix, mask)
				Bin.tag2 = {"truth":truth_matrix, "observed":observed, "mask":mask}
				ready_bins.append(Bin)

	return ready_bins

# get a set of bins with no missing data
def _filter_missing_data( bins, min_read_depth=1 ):
	cpg_bins_complete = _filter_bad_reads(bins)
	# secondary depth filter
	cpg_bins_complete_depth = [bin_ for bin_ in cpg_bins_complete if bin_.matrix.shape[0] >= min_read_depth]
	return cpg_bins_complete_depth





import pandas as pd
import numpy as np
import logging
import os
from clubcpg.ConnectToCpGNet import TrainWithPReLIM
from clubcpg.ParseBam import BamFileReadParser
from clubcpg_prelim import PReLIM
from pebble import ProcessPool
from joblib import load


class Imputation:
    """The class providing convienent APIs to train models and impute from models using PReLIM
    """

    def __init__(self, cpg_density: int, bam_file: str, mbias_read1_5=None, 
        mbias_read1_3=None, mbias_read2_5= None, mbias_read2_3=None, processes=-1):
        """[summary]
        
        Arguments:
            cpg_density {int} -- Number of CpGs this class instance will be used for
            bam_file {str} -- path to the bam file
        
        Keyword Arguments:
            mbias_read1_5 {[type]} -- [description] (default: {None})
            mbias_read1_3 {[type]} -- [description] (default: {None})
            mbias_read2_5 {[type]} -- [description] (default: {None})
            mbias_read2_3 {[type]} -- [description] (default: {None})
            processes {int} -- number or CPUs to use when parallelization can be utilized, default= All available (default: {-1})
        """

        self.cpg_density = cpg_density
        self.bam_file = bam_file
        self.mbias_read1_5 = mbias_read1_5
        self.mbias_read1_3 = mbias_read1_3
        self.mbias_read2_5 = mbias_read2_5
        self.mbias_read2_3 = mbias_read2_3
        self.processes = processes

    def extract_matrices(self, coverage_data_frame: pd.DataFrame, sample_limit: int = None, return_bins=False):
        """Extract CpG matrices from bam file.
        
        Arguments:
            coverage_data_frame {pd.DataFrame} -- Output of clubcpg-coverage read in as a csv file
        
        Keyword Arguments:
            return_bins {bool} -- Return the bin location along with the matrix (default: {False})
        
        Returns:
            [tuple] -- Returns tuple of (bin, np.array) if returns_bins = True else returns only np.array
        """

        subset = coverage_data_frame[coverage_data_frame['cpgs'] == self.cpg_density]
        bins_of_interest = subset['bin'].unique()

        # Downsample the training bins if requested and necessary
        if sample_limit and len(bins_of_interest) > sample_limit:
            bins_of_interest = np.random.choice(bins_of_interest, size=sample_limit)

        # Use the pebbel ProcessPool because it can handle hanging processes with a timeout
        complete_results = []
        with ProcessPool(max_workers=self.processes) as pool:
            future = pool.map(self._multiprocess_extract, bins_of_interest, timeout=5)

            iterator = future.result()

            while True:
                try:
                    result = next(iterator)
                    complete_results.append(result)
                except StopIteration:
                    break
                except TimeoutError as error:
                    print("Timeout caught - {}".format(error.args[1]))
                except Exception as error:
                    print("Unknown exception = {}".format(error))

        bins, matrices = zip(*complete_results)

        # destroy the pool
        pool.close()

        # Remove any potential bad data
        clean_matrices = []
        clean_bins = []
        for matrix, bin_ in zip(matrices, bins):
            try:
                if matrix.shape[1] == self.cpg_density:
                    clean_matrices.append(matrix)
                    clean_bins.append(bin_)
            except IndexError as e:
                logging.info("Index error at bin {}".format(bin_))
                logging.error(str(e))
                continue

        # if len(clean_matrices) > 0:
        #     clean_matrices = np.array(clean_matrices)

        clean_matrices = np.array(clean_matrices)

        if return_bins:
            return clean_bins, clean_matrices
        else:
            return clean_matrices


    def _multiprocess_extract(self, one_bin: str):
        """Function to be used for multiprocessing
        
        Arguments:
            one_bin {str} -- bin id as "chr7_222222"
        
        Returns:
            [tuple] -- bin, matrix
        """
        try:
            read_parser = BamFileReadParser(self.bam_file, 20, read1_5=self.mbias_read1_5, read1_3=self.mbias_read1_3, read2_5=self.mbias_read2_5, read2_3=self.mbias_read2_3)
            chrom, loc = one_bin.split("_")
            loc = int(loc)
            reads = read_parser.parse_reads(chrom, loc-100, loc) # TODO unhardcode bin size
            matrix = read_parser.create_matrix(reads)
            matrix = matrix.dropna(how="all")
            # if matrix.shape[0] == 0:
            #     return None
            matrix = matrix.fillna(-1)
            matrix = np.array(matrix)
            matrix = matrix.astype('int8')
        except: # BAD EXCEPTION
            return (one_bin, np.array([]))

        return (one_bin, matrix)


    def train_model(self, output_folder: str, matrices: iter):
        """Train a CpGNet model using :class:`.TrainWithCpGNet`
        
        Arguments:
            output_folder {str} -- Folder to save trained models
            matrices {iter} -- An iterable of CpGMatrices - ideally obtained through Imputation.extract_matrices()
        
        Returns:
            [keras model] -- Returns the trained CpGNet model
        """

        train_model = TrainWithPReLIM(cpg_density=self.cpg_density, save_path=output_folder)
        model = train_model.train_model(matrices)

        return model

    @staticmethod
    def postprocess_predictions(predicted_matrix):
        """Takes array with predicted values and rounds them to 0 or 1 if threshold is exceeded
        
        Arguments:
            predicted_matrix {[type]} -- matrix generated by imputation
        
        Returns:
            [type] -- predicted matrix predictions as 1, 0, or NaN
        """

        processed_array = []
        for array in predicted_matrix:
            new_array = []
            for item in array:
                if item != 1 and item != 0:
                    if item <= 0.2: #TODO un-hardcode this
                        new_array.append(0.0)
                    elif item >= 0.8: #TODO un-hardcode this
                        new_array.append(1.0)
                    else:
                        new_array.append(np.nan)
                else:
                    new_array.append(item)

            processed_array.append(new_array)

        return np.array(processed_array)




    def impute_from_model(self, models_folder: str, matrices: iter, postprocess=True):
        """Generator to provide imputed matrices on-the-fly
        
        Arguments:
            models_folder {str} -- Path to directory containing trained CpGNet models
            matrices {iter} -- An iterable containging n x m matrices with n=cpgs and m=reads
        
        Keyword Arguments:
            postprocess {bool} -- Round imputed values to 1s and 0s  (default: {True})
        """

        model_path = os.path.join(models_folder, "saved_model_{}_cpgs.prelim".format(self.cpg_density))

        trained_model = PReLIM(cpgDensity=self.cpg_density)
        print("Successfully loaded model: {}".format(model_path), flush=True)
        trained_model.model = load(model_path)

        for m in matrices:
            # only impute if there is an unknown
            if -1 in m:
                m = m.astype(float)
                pm = trained_model.impute(m)
                if postprocess:
                    pm = self.postprocess_predictions(pm)
            # Nothing to impute, passback original matrix to keep list in order
            else:
                pm = m.copy()
                
            # K.clear_session()
            yield pm   



from sklearn.cluster import DBSCAN
import pandas as pd
from pandas.core.indexes.base import InvalidIndexError
import numpy as np
import logging
import os
from clubcpg.ParseBam import BamFileReadParser
from clubcpg.OutputComparisonResults import OutputIndividualMatrixData
from clubcpg.Imputation import Imputation
import datetime
from multiprocessing import Pool
import time
import tempfile
from sklearn.utils import shuffle


class ClusterReads:
    """
    This class is used to take a dataframe or matrix of reads and cluster them

    :Example:
    >>> from clubcpg.ClusterReads import ClusterReads
    >>> cluster = ClusterReads(bam_a="/path/to/file.bam", bam_b="/path/to/file.bam", bins_file="/path/to/file.csv", suffix="chr19")
    >>> cluster.execute()

    """

    def __init__(self, bam_a: str, bam_b=None, bin_size=100, bins_file=None, output_directory=None, num_processors=1,
        cluster_member_min=4, read_depth_req=10, remove_noise=True, mbias_read1_5=None, 
        mbias_read1_3=None, mbias_read2_5=None, mbias_read2_3=None, suffix="", no_overlap=True, permute_labels=False):

        self.bam_a = bam_a
        self.bam_b = bam_b
        self.bin_size = int(bin_size)
        self.bins_file = bins_file
        self.output_directory = output_directory
        self.num_processors = num_processors
        self.cluster_member_min = cluster_member_min
        self.read_depth_req = read_depth_req
        self.remove_noise = remove_noise
        self.mbias_read1_5 = mbias_read1_5
        self.mbias_read1_3 = mbias_read1_3
        self.mbias_read2_5 = mbias_read2_5
        self.mbias_read2_3 = mbias_read2_3
        self.suffix = suffix
        self.no_overlap = no_overlap
        self.permute_labels = permute_labels
        
        if bam_b:
            self.single_file_mode = False
        else:
            self.single_file_mode = True

    # Remove clusters with less than n members
    def filter_data_frame(self, matrix: pd.DataFrame):
        """
        Takes a dataframe of clusters and removes any groups with less than self.cluster_member_min members

        :param matrix: dataframe of clustered reads
        :type: pd.DataFrame
        :return: input matrix with some clusters removed

        """
        output = matrix.copy()
        # duplicate indexes exist from concatenation, reset in the index to prevent dropping unintended rows
        output.reset_index(drop=True, inplace=True)
        for cluster in output['class'].unique():
            df = output[output['class'] == cluster]
            if len(df) < self.cluster_member_min:
                indexes = df.index
                output.drop(indexes, inplace=True)

        return output

    # Get only the matrices made up of reads from A OR B
    @staticmethod
    def get_unique_matrices(filtered_matrix):
        unique_dfs = []
        for label in filtered_matrix['class'].unique():
            df = filtered_matrix[filtered_matrix['class'] == label]
            if len(df['input'].unique()) == 1:  # A or B
                unique_dfs.append(df)

        return unique_dfs

    # Get matrices with reads made up of A AND B
    @staticmethod
    def get_common_matrices(filtered_matrix):
        shared_dfs = []
        for label in filtered_matrix['class'].unique():
            df = filtered_matrix[filtered_matrix['class'] == label]
            if len(df['input'].unique()) > 1:  # A and B
                shared_dfs.append(df)

        return shared_dfs

    # Get the means for all unique matrices
    def get_unique_means(self, filtered_matrix):
        output = []
        for matrix in self.get_unique_matrices(filtered_matrix):
            input_file = matrix['input'].unique()[0]
            matrix_mean = np.array(matrix)[:, :-2].mean()
            output.append((input_file, matrix_mean))

        return output

    # Get the means for all common matrices
    def get_common_means(self, filtered_matrix):
        output = []
        for matrix in self.get_common_matrices(filtered_matrix):
            matrix_mean = np.array(matrix)[:, :-2].mean()
            output.append(matrix_mean)

        return output

    # Generate a string label for each bin
    @staticmethod
    def make_bin_label(chromosome, stop_loc):
        return "_".join([chromosome, str(stop_loc)])

    @staticmethod
    def get_input_counts(df):
        output = {}
        for input_label in df['input'].unique():
            cts = len(df[df['input'] == input_label])
            output[input_label] = cts

        return output

    # Takes the output of process_bins() and converts it into list of lines of data for output
    def generate_individual_matrix_data(self, filtered_matrix, chromosome, bin_loc):
        """
        Take output of process_bins() and converts it into a list of lines of text data for output

        :param filtered_matrix: dataframe returned by :meth:`.ClusterReads.filter_data_frame`
        :param chromosome: chromosome as "Chr5"
        :param bin_loc: location representing the bin given as the end coordinate, ie 590000
        :type filtered_matrix: pd.DataFrame
        :type chromosome: string
        :type bin_loc: string
        :return: comma separated lines extracted from the filtered matrix, containing chromosome and bin info
        :rtype: list

        """
        # Individual comparisons data
        lines = []
        unique_groups = self.get_unique_matrices(filtered_matrix)
        common_groups = self.get_common_matrices(filtered_matrix)
        bin_label = self.make_bin_label(chromosome, bin_loc)

        for matrix in unique_groups:
            cpg_matrix = np.array(matrix.drop(['class', 'input'], axis=1))
            # get a semi-colon separated string of 1s and 0s representing the CpG pattern
            cpg_pattern = ";".join([str(int(x)) for x in list(cpg_matrix[0])])
            m_mean = cpg_matrix.mean()
            num_cpgs = cpg_matrix.shape[1]
            read_number = len(matrix)
            input_label = matrix['input'].unique()[0]
            class_label = matrix['class'].unique()[0]
            input_counts = self.get_input_counts(matrix)
            split_n_cpgs = ';'.join(["{}={}".format(x[0], x[1]) for x in input_counts.items()])

            out_line = ",".join([bin_label, input_label, str(m_mean), str(class_label), str(read_number),
                                str(num_cpgs), cpg_pattern, split_n_cpgs])
            lines.append(out_line)

        for matrix in common_groups:
            cpg_matrix = np.array(matrix.drop(['class', 'input'], axis=1))
            # get a semi-colon separated string of 1s and 0s representing the CpG pattern
            cpg_pattern = ";".join([str(int(x)) for x in list(cpg_matrix[0])])
            m_mean = cpg_matrix.mean()
            num_cpgs = cpg_matrix.shape[1]
            read_number = len(matrix)
            input_label = "".join(list(matrix['input'].unique()))
            class_label = matrix['class'].unique()[0]
            input_counts = self.get_input_counts(matrix)
            split_n_cpgs = ';'.join(["{}={}".format(x[0], x[1]) for x in input_counts.items()])
            out_line = ",".join([bin_label, input_label, str(m_mean), str(class_label), str(read_number),
                                str(num_cpgs), cpg_pattern, split_n_cpgs])
            lines.append(out_line)

        return lines

    @staticmethod
    def attempt_cpg_position_correction(reads, parser: BamFileReadParser):
        """
        Take the reads and a parser object, attempted cpg position correction and return corrected reads

        :param reads: parsed reads from BamFileReadParser
        :param parser: an instance of the BamFileReadParser object
        :return: reads with CpG positions corrected

        """

        corrected_reads = parser.correct_cpg_positions(reads)
        return corrected_reads

    # MAIN METHOD
    def process_bins(self, bin):
        """
        This is the main method and should be called using Pool.map It takes one bin location and uses the other helper
        functions to get the reads, form the matrix, cluster it with DBSCAN, and output the cluster data as text lines
        ready to writing to a file.

        :param bin: string in this format: "chr19_55555"
        :return: a list of lines representing the cluster data from that bin

        """
        try:
            chromosome, bin_loc = bin.split("_")
        except ValueError:
            return None
        bin_loc = int(bin_loc)

        # Create bam parser and parse reads
        bam_parser_A = BamFileReadParser(self.bam_a, 20, read1_5=self.mbias_read1_5, read1_3=self.mbias_read1_3,
                                        read2_5=self.mbias_read2_5, read2_3=self.mbias_read2_3, no_overlap=self.no_overlap)
        reads_A = bam_parser_A.parse_reads(chromosome, bin_loc - self.bin_size, bin_loc)

        if not self.single_file_mode:
            bam_parser_B = BamFileReadParser(self.bam_b, 20, read1_5=self.mbias_read1_5, read1_3=self.mbias_read1_3,
                                            read2_5=self.mbias_read2_5, read2_3=self.mbias_read2_3, no_overlap=self.no_overlap)
            reads_B = bam_parser_B.parse_reads(chromosome, bin_loc - self.bin_size, bin_loc)

        # This try/catch block returns None for a bin if any discrepancies in the data format of the bins are detected.
        # The Nones are filtered out during the output of the data
        try:
            # create matrix  drop NA
            # This matrix is actually a pandas dataframe
            matrix_A = bam_parser_A.create_matrix(reads_A).dropna()

            # Attempt to correct CpG Position if necessary
            if len(matrix_A) == 0:
                reads_A = self.attempt_cpg_position_correction(reads_A, bam_parser_A)
                matrix_A = bam_parser_A.create_matrix(reads_A).dropna()
            if not self.single_file_mode:
                matrix_B = bam_parser_B.create_matrix(reads_B).dropna()

                # attempt to correct CpG position in B if necessary
                if len(matrix_B) == 0:
                    reads_B = self.attempt_cpg_position_correction(reads_B, bam_parser_B)
                    matrix_B = bam_parser_B.create_matrix(reads_B).dropna()

        except ValueError as e:
            logging.error("ValueError when creating matrix at bin {}. Stack trace will be below if log level=DEBUG".format(bin))
            logging.debug(str(e))
            return None
        except InvalidIndexError as e:
            logging.error("Invalid Index error when creating matrices at bin {}".format(bin))
            logging.debug(str(e))
            return None

        # if read depths are still not a minimum, skip
        if matrix_A.shape[0] < self.read_depth_req:
            return None
        if not self.single_file_mode:
            if matrix_B.shape[0] < self.read_depth_req:
                return None

        # create labels and add to dataframe

        # If two files label each A and B, otherwise use file_name as label
        if not self.single_file_mode:
            try:
                labels_A = ['A'] * len(matrix_A)
                matrix_A['input'] = labels_A
                labels_B = ['B'] * len(matrix_B)
                matrix_B['input'] = labels_B
            except TypeError:
                logging.debug("TypeError when adding labels at bin {}".format(bin))
                return None
        else:
            labels_A = [os.path.basename(self.bam_a)] * len(matrix_A)
            matrix_A['input'] = labels_A

        if not self.single_file_mode:
            try:
                # ensure they have the same CpG positions
                matrix_B.columns = matrix_A.columns
                full_matrix = pd.concat([matrix_A, matrix_B], sort=False)
            except ValueError as e:
                logging.error("Matrix concat error in bin {}".format(bin))
                # logging.debug(str(e))
                return None
        else:
            full_matrix = matrix_A

        if self.permute_labels:
            # Randomly permute the input labels, has no effect in single file mode
            full_matrix['input'] = shuffle(full_matrix['input'].values)
            full_matrix = full_matrix.sort_values(by='input')

        # Get data without labels for clustering
        data_to_cluster = np.array(full_matrix)[:, :-1]

        # Create DBSCAN classifier and cluster add cluster classes to df
        clf = DBSCAN(min_samples=2)
        try:
            labels = clf.fit_predict(data_to_cluster)
        except ValueError as e:
            # log error
            logging.error("ValueError when trying to cluster bin {}".format(bin))
            logging.debug(str(e))
            return None

        full_matrix['class'] = labels

        # Filter out any clusters with less than a minimum
        filtered_matrix = self.filter_data_frame(full_matrix)
        if self.remove_noise:
            filtered_matrix = filtered_matrix[filtered_matrix['class'] != -1]

        # return generate_output_data(filtered_matrix, chromosome, bin_loc)
        return self.generate_individual_matrix_data(filtered_matrix, chromosome, bin_loc)

    def execute(self, return_only=False):
        """
        This method will start multiprocessing execution of this class.

        :param return_only: Whether to return the results as a variabel (True) or write to file (False)
        :type return_only: bool
        :return: list of lists if :attribute: `return_only` False otherwise None
        :rtype: list or None

        """
        start_time = datetime.datetime.now().strftime("%y-%m-%d")

        def track_progress(job, update_interval=60):
            while job._number_left > 0:
                logging.info("Tasks remaining = {0}".format(
                    job._number_left * job._chunksize
                ))
                time.sleep(update_interval)

        bins = []
        with open(self.bins_file, "r") as f:
            for line in f:
                data = line.strip().split(",")
                bins.append(data[0])

        pool = Pool(processes=self.num_processors)
        results = pool.map_async(self.process_bins, bins)
        track_progress(results)
        results = results.get()

        if return_only:
            return results

        else:
            output = OutputIndividualMatrixData(results)
            output.write_to_output(self.output_directory, "Clustering.{}{}.{}".format(os.path.basename(self.bam_a), self.suffix, start_time))


class ClusterReadsWithImputation(ClusterReads):
    """
    This class is used to perfom the same clustering, but also enabled the ability to perform imputation during clustering.
    This inherits from :class:`.ClusterReads`

    :Example:
    >>> from clubcpg.ClusterReads import ClusterReadsWithImputation
    >>> cluster = ClusterReadsWithImputation(...)
    >>> cluster.execute()

    """
    
    def __init__(self, bam_a: str, bam_b=None, bin_size=100, bins_file=None, output_directory=None, num_processors=1,
        cluster_member_min=4, read_depth_req=10, remove_noise=True, mbias_read1_5=None,
        mbias_read1_3=None, mbias_read2_5=None, mbias_read2_3=None, suffix="", no_overlap=True, models_A=None, models_B=None, chunksize=10000):

        self.models_A = models_A
        self.models_B = models_B
        self.chunksize = chunksize

        super().__init__(bam_a, bam_b, bin_size, bins_file, output_directory, 
        num_processors, cluster_member_min, read_depth_req, remove_noise, 
        mbias_read1_5, mbias_read1_3, mbias_read2_5, mbias_read2_3, suffix, no_overlap)

    def get_coverage_data(self, cpg_density=None):
        coverage_data = pd.read_csv(self.bins_file, header=None)
        coverage_data.columns = ['bin', 'reads', 'cpgs']

        return coverage_data

    def create_dictionary(self, bins, matrices):
        output = dict()
        for b, m in zip(bins, matrices):
            output[b] = m

        return output


    @staticmethod
    def filter_coverage_data(coverage_data, cpg_density):
        return coverage_data[coverage_data['cpgs'] == cpg_density]

    def execute(self, return_only=False):

        coverage_data = self.get_coverage_data()

        # start the main loop for imputation of these CpGs
        final_results_tf = tempfile.TemporaryFile(mode="w+t")
        final_results_tf.write("bin,input_label,methylation,class_label,read_number,cpg_number,cpg_pattern,class_split" + "\n")
        for i in range(2,6):
            print("Starting Imputation of CpG density {}...".format(i))

            imputer_A = Imputation(cpg_density=i, 
                bam_file=self.bam_a,
                mbias_read1_5=self.mbias_read1_5,
                mbias_read1_3=self.mbias_read1_3,
                mbias_read2_5=self.mbias_read2_5,
                mbias_read2_3=self.mbias_read2_3,
                processes=self.num_processors,
            )

            if self.bam_b:
                imputer_B = Imputation(cpg_density=i, 
                    bam_file=self.bam_b,
                    mbias_read1_5=self.mbias_read1_5,
                    mbias_read1_3=self.mbias_read1_3,
                    mbias_read2_5=self.mbias_read2_5,
                    mbias_read2_3=self.mbias_read2_3,
                    processes=self.num_processors
                )

            # Subset for CpG density
            sub_coverage_data = self.filter_coverage_data(coverage_data, i)

            # Split into chunks for memory management
            n = self.chunksize
            chunks = [sub_coverage_data[i * n:(i + 1) * n] for i in range((len(sub_coverage_data) + n - 1) // n )]
            n_chunks = len(chunks)
            print("Divided into {} chunks for processing...".format(n_chunks), flush=True)

            for j, chunk in enumerate(chunks):
                print("Extracting matrices from chunk {}/{}...".format(j+1,n_chunks))
                bins_A, matrices_A = imputer_A.extract_matrices(chunk, return_bins=True)

                if self.bam_b:
                    print("Extracting from second imput bam...", flush=True)
                    bins_B, matrices_B = imputer_B.extract_matrices(chunk, return_bins=True)
                else:
                    bins_B = None
                    matrices_B = None

                # Create dictionary with key as bin and values as a list like [matrix_A, matrix_B]
                data_A_dict = self.create_dictionary(bins_A, matrices_A)
                if self.bam_b:
                    data_B_dict = self.create_dictionary(bins_B, matrices_B)

                # Attempt to impute
                print("Imputing chunk {}/{}...".format(j+1, n_chunks), flush=True)
                imputed_matrices_A = imputer_A.impute_from_model(self.models_A, list(data_A_dict.values()))
                if self.bam_b:
                    imputed_matrices_B = imputer_B.impute_from_model(self.models_B, data_B_dict.values())
                else:
                    imputed_matrices_B = None

                data_imputed_A_dict = self.create_dictionary(data_A_dict.keys(), imputed_matrices_A)

                if self.bam_b:
                    data_imputed_B_dict = self.create_dictionary(data_B_dict.keys(), imputed_matrices_B)

                # Combine and cluster as normal
                # Maybe do this in the future
                # def _multiprocess_cluster_work(bin_):
                #     pass

                for bin_ in data_imputed_A_dict.keys():
                    matrix_A = data_imputed_A_dict[bin_]
                    matrix_A = pd.DataFrame(matrix_A)
                    matrix_A = matrix_A.dropna()
                    if matrix_A.shape[0] < self.read_depth_req:
                        continue
                    if self.bam_b:
                        try:
                            matrix_B = data_imputed_B_dict[bin_]
                        # matrix doesnt exist in other file
                        except KeyError:
                            logging.info("Covered bin {} doesnt exist in second file".format(bin_))
                            continue
                        matrix_B = pd.DataFrame(matrix_B)
                        matrix_B = matrix_B.dropna()
                        if matrix_B.shape[0] < self.read_depth_req:
                            # TODO SAME AS ABOVE
                            continue
                    
                    if self.bam_b:
                        labels_A = ['A'] * len(matrix_A)
                        matrix_A['input'] = labels_A
                        labels_B = ['B'] * len(matrix_B)
                        matrix_B['input'] = labels_B
                    else:
                        labels_A = [os.path.basename(self.bam_a)] * len(matrix_A)
                        matrix_A['input'] = labels_A

                    if self.bam_b:
                        try:
                            full_matrix = pd.concat([matrix_A, matrix_B])
                        except ValueError as e:
                            logging.error("Matrix concat eror in bin {}".format(bin_))
                            continue
                    else:
                        full_matrix = matrix_A

                    # get data to cluster
                    data_to_cluster = np.array(full_matrix)[:,:-1]
                    clf = DBSCAN(min_samples=2)
                    try:
                        labels = clf.fit_predict(data_to_cluster)
                    except ValueError as e:
                        logging.error("ValueError when trying to cluster bin {}".format(bin_))
                        continue

                    full_matrix['class'] = labels

                    filtered_matrix = super().filter_data_frame(full_matrix)
                    if self.remove_noise:
                        filtered_matrix = filtered_matrix[filtered_matrix['class'] != -1]
                    
                    # generate output lines
                    chromosome, bin_loc = bin_.split("_")
                    output_lines = super().generate_individual_matrix_data(filtered_matrix, chromosome, bin_loc)
                    for line in output_lines:
                        final_results_tf.write(line+"\n")

        # CLUSTER ALL OTHER BINS LIKE NORMAL WITHOUT IMPUTATION
        print("Performing clustering on the rest of the bins with no imputaiton...", flush=True)
        unimputable_coverage = coverage_data[coverage_data['cpgs'] >= 6]

        unimputable_temp = tempfile.NamedTemporaryFile(mode="w+t")
        unimputable_file_name = unimputable_temp.name
        logging.info("Temp file for unimputable bins is {}".format(unimputable_file_name))
        logging.info("unimputable bins looks like this...")
        logging.info(unimputable_coverage.head())

        unimputable_coverage.to_csv(unimputable_file_name, header=False, index=False)


        cluster_reads = ClusterReads(
            bam_a=self.bam_a,
            bam_b=self.bam_b,
            bin_size=self.bin_size,
            bins_file=unimputable_file_name,
            output_directory=self.output_directory,
            num_processors=self.num_processors,
            cluster_member_min=self.cluster_member_min,
            read_depth_req=self.read_depth_req,
            remove_noise=self.remove_noise,
            mbias_read1_5=self.mbias_read1_5,
            mbias_read1_3=self.mbias_read1_3,
            mbias_read2_5=self.mbias_read2_5,
            mbias_read2_3=self.mbias_read2_3,
            suffix=self.suffix,
            no_overlap=self.no_overlap
        )

        # Write this output to the output temp file
        results = cluster_reads.execute(return_only=True)
        for result in results:
            if result:
                for line in result:
                    final_results_tf.write(line + "\n")

        # rewrite the tempfile to a final output file
        final_results_tf.seek(0)
        # output = 'output_dir/basename_suffix_cluster_results.csv'
        output_file = os.path.join(self.output_directory, os.path.basename(self.bam_a) + self.suffix + "_cluster_results.csv")
        with open(output_file, "w") as final:
            for line in final_results_tf:
                final.write(line)

        unimputable_temp.close()



import pysam
import pandas as pd
from collections import defaultdict
import logging
import re


class BamFileReadParser:
    """
    Used to simplify the opening and reading from BAM files. BAMs must be coordinate sorted and indexed.

    :Example:
        >>> from clubcpg.ParseBam import BamFileReadParser
        >>> parser = BamFileReadParser("/path/to/data.BAM", quality_score=20, read1_5=3, read1_3=4, read2_5=7, read2_3=1)
        >>> reads = parser.parse_reads("chr7", 10000, 101000)
        >>> reads = parser.correct_cpg_positions(reads) # This step is optional, but highly recommended
        >>> matrix = parser.create_matrix(reads)


    """

    def __init__(self, bamfile, quality_score, read1_5=None, read1_3=None,
                 read2_5=None, read2_3=None, no_overlap=True):
        """
        Class used to read WGBSeq reads from a BAM file, extract methylation, and convert into data frame

        :param bamfile: Path to bam file location
        :param quality_score: Only include reads >= this fastq quality
        :param read1_5: mbias ignore read1 5'
        :param read1_3: mbias ignore read1 3'
        :param read2_5: mbias ignore read2 5'
        :param read2_3: mbias ignore read2 3'
        :param no_overlap: bool. If overlap exists between two reads, ignore that region from read 2.
        """

        self.mapping_quality = quality_score
        self.bamfile = bamfile
        self.read1_5 = read1_5
        self.read1_3 = read1_3
        self.read2_5 = read2_5
        self.read2_3 = read2_3
        self.full_reads = []
        self.read_cpgs = []
        self.no_overlap = no_overlap

        if read1_5 or read2_5 or read1_3 or read2_3:
            self.mbias_filtering = True
        else:
            self.mbias_filtering = False

        self.OpenBamFile = pysam.AlignmentFile(bamfile, 'rb')
        # Check for presence of index file
        index_present = self.OpenBamFile.check_index()
        if not index_present:
            raise FileNotFoundError("BAM file index is not found. Please create it using samtools index")

    # From open bam file, get locaiton of first read from the provided chromosome
    def get_location_of_first_read(self, chromosome):

        # Get reference lenghts
        ref_lens = dict(zip(self.OpenBamFile.references, self.OpenBamFile.lengths))

        for read in self.OpenBamFile.fetch(chromosome, 0, ref_lens[chromosome]):
            reads_start_loc = read.reference_start
            break

        return reads_start_loc

    # Get reads from the bam file, extract methylation state
    def parse_reads(self, chromosome: str, start:int , stop: int):
        """
        :param chromosome: chromosome as "chr6"
        :param start: start coordinate
        :param stop: end coordinate
        :return: List of reads and their positional tags as assigned by bismark
        """
        reads = []
        for read in self.OpenBamFile.fetch(chromosome, start, stop):
            if read.mapping_quality >= self.mapping_quality:
                reads.append(read)

        ## CIGAR FILTERING BY C. COARFA
        read_cpgs = []
        self.skipped_reads = set()

        read_index = -1
        self.query_count_hash = {}
        for read in reads:
            read_index +=1

            if not (read.query_name in self.query_count_hash):
                self.query_count_hash[read.query_name]=0
            
            self.query_count_hash[read.query_name] += 1
            # if (self.query_count_hash[read.query_name]>2):
            #     logging.info("Found read with more than 2 mappings: %s --> %s\n"%(read.query_name, self.query_count_hash[read.query_name]))
            
            # need to check for regular expression though
            no_indel_mapping = re.match("^\d+M$", read.cigarstring)
            if (no_indel_mapping):

                reduced_read = []
                # Join EVERY XM tag with its aligned_pair location
                for pair, tag in zip(read.get_aligned_pairs(), read.get_tag('XM')):
                    if pair[1]:
                        if read.flag == 83 or read.flag == 163 or read.flag == 16:
                            reduced_read.append((pair[1] - 1, tag))
                        else:
                            reduced_read.append((pair[1], tag))
                    else:
                        continue
    
                # if MBIAS was set, slice the joined list
                if self.mbias_filtering:
                    if read.is_read1:
                        mbias_5_prime = self.read1_5
                        # note taking the NEGATIVE of the value for the 3-prime
                        mbias_3_prime = -self.read1_3
                        if mbias_3_prime == 0:
                            mbias_3_prime = None
                        reduced_read = reduced_read[mbias_5_prime:mbias_3_prime]
                    if read.is_read2:
                        mbias_5_prime = self.read2_5
                        mbias_3_prime = -self.read2_3
                        if mbias_3_prime == 0:
                            mbias_3_prime = None
                        reduced_read = reduced_read[mbias_5_prime:mbias_3_prime]
                    
                read_cpgs.append(reduced_read)
            else:
                self.skipped_reads.add(read.query_name)

        self.full_reads = reads
        self.read_cpgs = read_cpgs

        # Correct overlapping paired reads if set, this is default behavior
        if self.no_overlap:
            try:
                read_cpgs = self.fix_read_overlap(reads, read_cpgs)
            except AttributeError:
                pass
                # print("Could not determine read 1 or 2. {}:{}-{}".format(chromosome, start, stop))
                # sys.stdout.flush()

        # Filter the list for positions between start-stop and CpG (Z/z) tags
        output = []
        read_cpg_index = -1
        
        found_cpg_count = 0
        
        for read_cpg in read_cpgs:
            read_cpg_index +=1
            temp = []
            for pos, tag in read_cpg:
                if pos and (pos > start) and (pos <= stop) and ((tag == 'Z') or (tag == 'z')):
                    temp.append((pos, tag))
                    found_cpg_count += 1
                    
            output.append(temp)
        return output

    def create_matrix(self, read_cpgs):
        """
        Converted parsed reads into a pandas dataframe.

        :param read_cpgs: read CpGs generated by self.parse_reads
        :type read_cpgs: iterable

        :return: matrix methylated (1) and unmethylated (0) states
        :rtype: pd.DataFrame

        """
        series = []
        data_index = -1
        for data in read_cpgs:
            data_index += 1
            positions = []
            statues = []
            num_positions = 0
            dup_flag = False
            positions_set = set()
            for pos, status in data:
                num_positions += 1
                if pos in positions_set:
                    dup_flag = True
                positions_set.add(pos)
                
                positions.append(pos)
                statues.append(status)
            

            if dup_flag:
                pass
            else:
                if num_positions > 0:
                    series.append(pd.Series(statues, positions))

        try:
            matrix = pd.concat(series, axis=1, ignore_index=True)
        except BaseException as e:
            raise ValueError("Empty matrix")

        matrix = matrix.replace('Z', 1)
        matrix = matrix.replace('z', 0)

        return matrix.T

    def fix_read_overlap(self, full_reads, read_cpgs):
        """Takes pysam reads and read_cpgs generated during parse reads and removes any
        overlap between read1 and read2. If possible it also stitches read1 and read2 together to create
        a super read.

        :param full_reads: set of reads generated by self.parse_reads()
        :param read_cpgs: todoo
        :return: A list in the same format as read_cpgs input, but corrected for paired read overlap
        """
        # data for return
        fixed_read_cpgs = []
        # Combine raw reads and extracted tags
        combined = []
        for read, state in zip(full_reads, read_cpgs):
            combined.append((read, state))

        # Get names of all the reads present
        # query_names = [x.query_name for x in full_reads]
        query_names = []
        for x in full_reads:
            if (not (x.query_name in self.skipped_reads)) and (self.query_count_hash[x.query_name]<=2):
                query_names.append(x.query_name)

        # Match paired reads by query_name
        tally = defaultdict(list)
        for i, item in enumerate(query_names):
            tally[item].append(i)

        for key, value in sorted(tally.items()):
            # A pair exists, process it
            if len(value) == 2:
                # Set read1 and read2 correctly
                if combined[value[0]][0].is_read1:
                    read1 = combined[value[0]]
                    read2 = combined[value[1]]

                elif combined[value[1]][0].is_read1:
                    read1 = combined[value[1]]
                    read2 = combined[value[0]]

                # both reads have same value, this shouldn't be. Drop one completely, dont
                # bother with overlap
                elif combined[value[0]][0].is_read1 == combined[value[1]][0].is_read1:
                    fixed_read_cpgs.append(combined[value[0]][1])
                    continue


                else:
                    raise AttributeError("Could not determine read 1 or read 2")

                # Find amount of overlap
                amount_overlap = 0
                r1_bps = [x[0] for x in read1[1]]
                r2_bps = [x[0] for x in read2[1]]

                if min(r1_bps) < min(r2_bps):
                    trim_direction = 5
                    for bp in r2_bps:
                        if bp and bp in r1_bps:
                            amount_overlap += 1
                else:
                    trim_direction = 3
                    for bp in r1_bps:
                        if bp and bp in r2_bps:
                            amount_overlap += 1

                # remove the overlap by trimming or discarding
                if amount_overlap == len(read2[1]):
                    # discard read 2, only append read 1
                    fixed_read_cpgs.append(read1[1])
                else:
                    # trim overlap
                    if trim_direction == 5:
                        new_read2_cpgs = read2[1][amount_overlap:]
                    elif trim_direction == 3:
                        new_read2_cpgs = read2[1][:-amount_overlap]
                    # stitch together read1 and read2
                    read1[1].extend(new_read2_cpgs)
                    fixed_read_cpgs.append(read1[1])

            elif len(value) == 1:
                # No pair, add to output
                fixed_read_cpgs.append(combined[value[0]][1])

        return fixed_read_cpgs

    @staticmethod
    def correct_cpg_positions(output: list):
        """
        For some reason, Bismark alignment produces instances where a CpG site location is incorrect by 1 bp, even
        after accounting for DNA strand alignmment. This function fixes this. If two cpgs have positions such as 4, 5
        (which is impossible because there needs to by a G between them) this function will convert all 5s to 4s. This
        only needs to be applied to matrices which are empty after dropna() is called.

        :param output: a list of lists of tuples. The output of self.parse_reads()

        :return: list of the same style, execpt the first position in the tuple will have a corrected CpG position.

        """
        # find all cpg positions
        cpg_positions = []
        for item in output:
            if item:
                for cpg in item:
                    cpg_positions.append(cpg[0])
        cpg_positions = sorted(list(set(cpg_positions)))

        # determine corrections
        corrections = {}
        for x in range(len(cpg_positions)):
            try:
                if cpg_positions[x + 1] == cpg_positions[x] + 1:
                    corrections[cpg_positions[x + 1]] = cpg_positions[x]
            except IndexError:  # end of cpg position list
                pass

        # correct items
        corrected_output = []
        for item in output:
            corrected_item = []
            if item:
                for cpg in item:
                    if cpg[0] in corrections.keys():
                        new_cpg = (corrections[cpg[0]], cpg[1])
                        corrected_item.append(new_cpg)
                    else:
                        corrected_item.append(cpg)
                corrected_output.append(corrected_item)
            else:
                corrected_output.append(item)

        return corrected_output






import os


# This class will be called by clubcpg-cluster to output its multiprocessing results output into csv files

class OutputIndividualMatrixData:

    def __init__(self, results):
        """
        :param results: Results objected generated by CompareClusterCompare.py's Pool.map of process_bins
        """
        self.results = results

    def write_to_output(self, filepath=None, prefix=None):
        """
        :param prefix: Prefix to name the output files
        :param filepath: Path to save the output files
        :return: None
        """

        if prefix and filepath:
            output_comparisons = open("{}_matrix_data.csv".format(os.path.join(filepath, prefix)), "w")

            output_comparisons.write("bin,input_label,methylation,class_label,read_number,cpg_number,cpg_pattern,class_split\n")

            for result in self.results:
                if result:
                    for line in result:
                        output_comparisons.write(line + "\n")
                else:
                    continue

            output_comparisons.close()

            return




from clubcpg_prelim import PReLIM
import os
from joblib import dump


class TrainWithPReLIM:
    """
    Used to train models using CpGnet
    """

    def __init__(self, cpg_density=None, save_path=None):
        """
        Class to train a CpGNet model from input data

        :param cpg_density: Number of CpGs
        :type cpg_density: int
        :param save_path: Location of folder to save the resulting model files. One per cpg density
        """
        if not cpg_density:
            raise AttributeError("CpG density must be specified")
        if not save_path:
            raise AttributeError("Folder to save trained model must be specified")
        self.save_path = save_path
        self.cpg_density = cpg_density
        self.model = PReLIM(cpgDensity=cpg_density)

    def save_net(self, model):
        """
        Save the network to a file

        :param model: The trained PReLIM model. Located at PReLIM.model
        :type model: :class:`clubcpg_prelim.PReLIM`
        :return: Path to the saved model
        """
        file_name = "saved_model_{}_cpgs.prelim".format(self.cpg_density)
        output = os.path.join(self.save_path, file_name)
        dump(model, output)
        print("Saved {} cpg model to {}".format(self.cpg_density, output))

        return output


    def train_model(self, bins: iter):
        """
        Train the CpGNet model on a list of provided bins

        :param bins: iterable containing CpG matrices of 1 (methylated), 0 (unmethylated), and -1 (unknown)
        :return: Path to the saved model file
        """
        self.model.train(bins, model_file="no")
        output = self.save_net(self.model.model)

        return output




from setuptools import setup

with open("README.md", "rt") as f:
    long_description = f.read()

setup(name="clubcpg",
      version="0.2.4",
      description="CluBCpG is a software package built to analyze whole genome bisulfite sequencing (WGBS) data",
      long_description=long_description,
      long_description_content_type="text/markdown",
      author="C. Anthony Scott, PhD",
      author_email="charles.scott@bcm.edu",
      url="https://github.com/waterlandlab/CluBCpG",
      license='MIT',
      packages=['clubcpg', 'clubcpg_prelim'],

      install_requires=[
          'pysam==0.15.2',
          'numpy==1.16.4',
          'scikit-learn==0.21.2',
          'joblib',
          'scipy', 
          'pandas==0.25.0',
          'fastcluster', 
          'pebble', 
          'tqdm'
      ],

      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: POSIX",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
      ],

      scripts=[
          'bin/clubcpg-coverage',
            'bin/clubcpg-cluster',
            'bin/clubcpg-impute-train',
            'bin/clubcpg-impute-coverage',
            'bin/clubcpg-impute-cluster',
      ]

      )
